//=============================================================================
// Servo.cpp - This file contains servo support that is based on the Arduino
//		servo subsystem.  It will later be extended to add some of the additional
//		functionality of the BAP HServo system.  Also may be a cross with 
//		what the SSC-32 also supporets
//==============================================================================

/*
 Servo.cpp - Interrupt driven Servo library for Arduino using 16 bit timers- Version 2
 Copyright (c) 2009 Michael Margolis.  All right reserved.
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.
 
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/* 
 
 A servo is activated by creating an instance of the Servo class passing the desired pin to the attach() method.
 The servos are pulsed in the background using the value most recently written using the write() method
 
 Note that analogWrite of PWM on pins associated with the timer are disabled when the first servo is attached.
 Timers are seized as needed in groups of 12 servos - 24 servos use two timers, 48 servos will use four.
 
 The methods are:
 
 Servo - Class for manipulating servo motors connected to Arduino pins.
 
 attach(pin )  - Attaches a servo motor to an i/o pin.
 attach(pin, min, max  ) - Attaches to a pin setting min and max values in microseconds
 default min is 544, max is 2400  
 
 write()     - Sets the servo angle in degrees.  (invalid angle that is valid as pulse in microseconds is treated as microseconds)
 writeMicroseconds() - Sets the servo pulse width in microseconds 
 read()      - Gets the last written servo pulse width as an angle between 0 and 180. 
 readMicroseconds()   - Gets the last written servo pulse width in microseconds. (was read_us() in first release)
 attached()  - Returns true if there is a servo attached. 
 detach()    - Stops an attached servos from pulsing its i/o pin. 
 
*/

#include "baplib.h"
#include "Servo.h"
#define clockCyclesPerMicrosecond 16								// We are running at 16mhz
#define usToTicks(_us)    (unsigned int)(((unsigned long) clockCyclesPerMicrosecond * _us) / 8)     // converts microseconds to tick (assumes prescale of 8)  // 12 Aug 2009
//#define ticksToUs(_ticks) (( (unsigned long)_ticks * 8)/ clockCyclesPerMicrosecond ) // converts from ticks back to microseconds
#define ticksToUs(_ticks) (( (unsigned long)_ticks * 8)/ (unsigned long)clockCyclesPerMicrosecond ) // converts from ticks back to microseconds



#define TRIM_DURATION       2                               // compensation ticks to trim adjust for digitalWrite delays // 12 August 2009


static servo_t servos[MAX_SERVOS];                          // static array of servo structures
static volatile char Channel[_Nbr_16timers ];             // counter for the servo being pulsed for each timer (or -1 if refresh interval)

u8 ServoCount = 0;                                     // the total number of attached servos


// convenience macros
#define SERVO_INDEX_TO_TIMER(_servo_nbr) ((u8)(_servo_nbr / SERVOS_PER_TIMER)) // returns the timer controlling this servo
#define SERVO_INDEX_TO_CHANNEL(_servo_nbr) (_servo_nbr % SERVOS_PER_TIMER)       // returns the index of the servo on this timer
#define SERVO_INDEX(_timer,_channel)  ((_timer*SERVOS_PER_TIMER) + _channel)     // macro to access servo index by timer and channel
#define SERVO(_timer,_channel)  (servos[SERVO_INDEX(_timer,_channel)])            // macro to access servo class by timer and channel

#define SERVO_MIN() (MIN_PULSE_WIDTH - this->min * 4)  // minimum value in uS for this servo
#define SERVO_MAX() (MAX_PULSE_WIDTH - this->max * 4)  // maximum value in uS for this servo 

/************ static functions common to all instances ***********************/
// Handle each of the channels for the WTimer.  Question is, do I reset the timer for each cycle or let it run
// and overflow.  Also we interrupt for each defined servo even if the servo is not currently active.  May want
// to fix that... - Also all active channels will trigger at the same time at the REFRESH_INTERAL.  Probably not
// a problem to have them all reset the TCNT to zero.

static __inline__ void handle_timerw_channel(u8 timer, volatile unsigned int* pGRx) __attribute__ ((always_inline));
static __inline__ void handle_timerw_channel(u8 timer, volatile unsigned int* pGRx)
{
	if( Channel[timer] == 0xff )
		TW.TCNT = 0; // channel set to -1 indicated that refresh interval completed so reset the timer 
	else
	{
		if( SERVO_INDEX(timer,Channel[timer]) < ServoCount && SERVO(timer,Channel[timer]).Pin.isActive == true )  
	  		low(SERVO(timer,Channel[timer]).Pin.nbr); // pulse this channel low if activated   
	}
	
	Channel[timer]++;    // increment to the next channel
	if( SERVO_INDEX(timer,Channel[timer]) < ServoCount && Channel[timer] < SERVOS_PER_TIMER) 
	{
		*pGRx = TW.TCNT + SERVO(timer,Channel[timer]).ticks;
		if(SERVO(timer,Channel[timer]).Pin.isActive == true)     // check if activated
  			high(SERVO(timer,Channel[timer]).Pin.nbr); 					// its an active channel so pulse it high   
	}  
	else 
	{ 
		// finished all channels so wait for the refresh period to expire before starting over 
		if( TW.TCNT <  (usToTicks(REFRESH_INTERVAL) + 4) )  // allow a few ticks to ensure the next OCR1A not missed
  			*pGRx = (unsigned int)usToTicks(REFRESH_INTERVAL);  
		else 
  			*pGRx = TW.TCNT + 4;  // at least REFRESH_INTERVAL has elapsed
		Channel[timer] = 0xff; // this will get incremented at the end of the refresh period to start again at the first channel
	}
}



extern "C" void TIMERW() __attribute__ ((interrupt_handler));
extern "C" void TIMERW()
{
	// Lets see what we woke up for.
	if (TW.TSRW.BIT.IMFA)	// GRA Match?
	{
		handle_timerw_channel(0, &TW.GRA);
		TW.TSRW.BIT.IMFA = 0;
	}
	if (TW.TSRW.BIT.IMFB)	// GRB Match?
	{
		handle_timerw_channel(1, &TW.GRB); 
		TW.TSRW.BIT.IMFB = 0;
	}
	if (TW.TSRW.BIT.IMFC)	// GRC Match?
	{
		handle_timerw_channel(2, &TW.GRC); 
		TW.TSRW.BIT.IMFC = 0;
	}
	if (TW.TSRW.BIT.IMFD)	// GRD Match?
	{
		handle_timerw_channel(3, &TW.GRD); 
		TW.TSRW.BIT.IMFD = 0;
	}
}


static void initTimerChannel(u8 iChan)
{  
	// Make sure the timer is setup to count on every 8 clocks, plus that the
	// Group index for the specified servo is initialized and the interrupt
	// assocated with the channel is enabled
	TW.TCRW.BIT.CKS = 0x3;		// clock / 8
	TW.TMRW.BIT.CTS = 1;		// start the clock;
	
	Channel[iChan] = 0xff; // this will get incremented at the end of the refresh period to start again at the first channel

	switch (iChan)
	{
		case 0:
  			TW.GRA  = (unsigned int)(usToTicks(REFRESH_INTERVAL));  
			TW.TSRW.BIT.IMFA = 0;		// clear the compare match value
			TW.TIERW.BIT.IMIEA = 1;		// enable the interrupt
			TW.TIOR0.BIT.IOA = 0;		// output match - don't change any io pins
			break;
		case 1:
  			TW.GRB  = (unsigned int)usToTicks(REFRESH_INTERVAL);  
			TW.TSRW.BIT.IMFB = 0;		// clear the compare match value
			TW.TIERW.BIT.IMIEB = 1;		// enable the interrupt
			TW.TIOR0.BIT.IOB = 0;		// output match - don't change any io pins
			break;
		case 2:
  			TW.GRC  = (unsigned int)usToTicks(REFRESH_INTERVAL);  
			TW.TSRW.BIT.IMFC = 0;		// clear the compare match value
			TW.TIERW.BIT.IMIEC = 1;		// enable the interrupt
			TW.TIOR1.BIT.IOC = 0;		// output match - don't change any io pins
			break;
		case 3:
  			TW.GRD  = (unsigned int)usToTicks(REFRESH_INTERVAL);  
			TW.TSRW.BIT.IMFD = 0;		// clear the compare match value
			TW.TIERW.BIT.IMIED = 1;		// enable the interrupt
			TW.TIOR1.BIT.IOD = 0;		// output match - don't change any io pins
	}
} 

static void finISR(u8 iChan)
{
	switch (iChan)
	{
		case 0:
  			TW.GRA  = (unsigned int)0xff;
			TW.TSRW.BIT.IMFA = 0;		// clear the compare match value
			TW.TIERW.BIT.IMIEA = 0;		// disable the interrupt
			break;
		case 1:
  			TW.GRB  = (unsigned int)0xff;  
			TW.TSRW.BIT.IMFB = 0;		// clear the compare match value
			TW.TIERW.BIT.IMIEB = 0;		// disable the interrupt
			break;
		case 2:
  			TW.GRC  = (unsigned int)0xff;  
			TW.TSRW.BIT.IMFC = 0;		// clear the compare match value
			TW.TIERW.BIT.IMIEC = 0;		// disable the interrupt
			break;
		case 3:
  			TW.GRD  = (unsigned int)0xff;  
			TW.TSRW.BIT.IMFD = 0;		// clear the compare match value
			TW.TIERW.BIT.IMIED = 0;		// disable the interrupt
	}
}

static u8 isTimerActive(u8 iChan)
{
  // returns true if any servo is active on this timer
  u8 channel;
  for(channel=0; channel < SERVOS_PER_TIMER; channel++) {
    if(SERVO(iChan,channel).Pin.isActive == true)
      return true;
  }
  return false;
}


/****************** end of static functions ******************************/

Servo::Servo()
{
  if( ServoCount < MAX_SERVOS) {
    this->servoIndex = ServoCount++;                    // assign a servo index to this instance
	servos[this->servoIndex].ticks = usToTicks(DEFAULT_PULSE_WIDTH);   // store default values  - 12 Aug 2009
  }
  else
    this->servoIndex = INVALID_SERVO ;  // too many servos 
}

u8 Servo::attach(int pin)
{
  return this->attach(pin, MIN_PULSE_WIDTH, MAX_PULSE_WIDTH);
}

u8 Servo::attach(int pin, int min, int max)
{
  if(this->servoIndex < MAX_SERVOS ) {
    pinMode( pin, OUTPUT) ;                                   // set servo pin to output
    servos[this->servoIndex].Pin.nbr = pin;  
    // todo min/max check: abs(min - MIN_PULSE_WIDTH) /4 < 128 
    this->min  = (MIN_PULSE_WIDTH - min)/4; //resolution of min/max is 4 uS
    this->max  = (MAX_PULSE_WIDTH - max)/4; 
    // initialize the timer if it has not already been initialized 
    u8 timer = SERVO_INDEX_TO_TIMER(servoIndex);
    if(isTimerActive(timer) == false)
      initTimerChannel(timer);    
    servos[this->servoIndex].Pin.isActive = true;  // this must be set after the check for isTimerActive
  } 
  return this->servoIndex ;
}

void Servo::detach()  
{
  servos[this->servoIndex].Pin.isActive = false;  
  u8 timer = SERVO_INDEX_TO_TIMER(servoIndex);
  if(isTimerActive(timer) == false) {
    finISR(timer);
  }
}

long map(long x, long in_min, long in_max, long out_min, long out_max)
{
  return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}

void Servo::write(int value)
{  
  if(value < MIN_PULSE_WIDTH)
  {  // treat values less than 544 as angles in degrees (valid values in microseconds are handled as microseconds)
    if(value < 0) value = 0;
    if(value > 180) value = 180;
    value = map(value, 0, 180, SERVO_MIN(),  SERVO_MAX());      
  }
  this->writeMicroseconds(value);
}

void Servo::writeMicroseconds(int value)
{
  // calculate and store the values for the given channel
  byte channel = this->servoIndex;
  if( (channel != INVALID_SERVO) && (channel < MAX_SERVOS) )   // ensure channel is valid
  {  
    if( value < SERVO_MIN() )          // ensure pulse width is valid
      value = SERVO_MIN();
    else if( value > SERVO_MAX() )
      value = SERVO_MAX();   
    
  	value = value - TRIM_DURATION;
    value = usToTicks(value);  // convert to ticks after compensating for interrupt overhead - 12 Aug 2009

    //u8 oldSREG = SREG;
    //cli();
    servos[channel].ticks = value;  
    //SREG = oldSREG;   
  } 
}

int Servo::read() // return the value as degrees
{
  return  map( this->readMicroseconds()+1, SERVO_MIN(), SERVO_MAX(), 0, 180);     
}

int Servo::readMicroseconds()
{
  unsigned int pulsewidth;
  if( this->servoIndex != INVALID_SERVO )
    pulsewidth = ticksToUs(servos[this->servoIndex].ticks)  + TRIM_DURATION ;   // 12 aug 2009
  else 
    pulsewidth  = 0;

  return pulsewidth;   
}

bool Servo::attached()
{
  return servos[this->servoIndex].Pin.isActive ;
}
